@@ -25,6 +25,7 @@
 #include "xdg-desktop-portal-dbus.h"
 #include "shell-dbus.h"
 
+#include "clipboard.h"
 #include "remotedesktop.h"
 #include "remotedesktopdialog.h"
 #include "screencastdialog.h"
@@ -34,6 +35,9 @@
 #include "session.h"
 #include "utils.h"
 
+#define RESTORE_FORMAT_VERSION 1
+#define REMOTE_DESKTOP_RESTORE_VARIANT_TYPE "(xxub" SCREEN_CAST_STREAMS_VARIANT_TYPE ")"
+
 typedef enum _GnomeRemoteDesktopDeviceType
 {
   GNOME_REMOTE_DESKTOP_DEVICE_TYPE_KEYBOARD = 1 << 0,
@@ -68,9 +72,25 @@ typedef struct _RemoteDesktopSession
   } select;
 
   struct {
+    gboolean persist_allowed;
     RemoteDesktopDeviceType device_types;
   } shared;
 
+  struct
+  {
+    gboolean clipboard_enabled;
+    gboolean clipboard_requested;
+    gulong selection_owner_changed_handler_id;
+    gulong selection_transfer_handler_id;
+  } clipboard;
+
+  RemoteDesktopPersistMode persist_mode;
+  GPtrArray *streams_to_restore;
+  struct {
+    GVariant *data;
+    int64_t creation_time;
+  } restored;
+
   GDBusMethodInvocation *start_invocation;
   RemoteDesktopDialogHandle *dialog_handle;
 } RemoteDesktopSession;
@@ -80,6 +100,16 @@ typedef struct _RemoteDesktopSessionClass
   SessionClass parent_class;
 } RemoteDesktopSessionClass;
 
+enum
+{
+  CLIPBOARD_SELECTION_OWNER_CHANGED,
+  CLIPBOARD_SELECTION_TRANSFER,
+
+  N_SIGNAL
+};
+
+static guint signals[N_SIGNAL];
+
 typedef struct _RemoteDesktopDialogHandle
 {
   Request *request;
@@ -107,6 +137,8 @@ static gboolean
 start_session (RemoteDesktopSession     *session,
                RemoteDesktopDeviceType   device_types,
                GPtrArray                *streams,
+               gboolean                  clipboard_enabled,
+               gboolean                  persist_allowed,
                GError                  **error);
 
 static void
@@ -127,6 +159,43 @@ remote_desktop_session_sources_selected (RemoteDesktopSession *session,
   session->select.screen_cast = *selection;
 }
 
+static void
+on_selection_owner_changed (OrgGnomeMutterRemoteDesktopSession *object,
+                            GVariant *arg_options,
+                            RemoteDesktopSession *session)
+{
+  g_signal_emit (session, signals[CLIPBOARD_SELECTION_OWNER_CHANGED], 0,
+                 arg_options);
+}
+
+static void
+on_selection_transfer (OrgGnomeMutterRemoteDesktopSession *object,
+                       const gchar *arg_mime_type,
+                       guint arg_serial,
+                       RemoteDesktopSession *session)
+{
+  g_signal_emit (session, signals[CLIPBOARD_SELECTION_TRANSFER], 0,
+                 arg_mime_type, arg_serial);
+}
+
+void
+remote_desktop_session_request_clipboard (RemoteDesktopSession *session)
+{
+  session->clipboard.clipboard_requested = TRUE;
+}
+
+gboolean
+remote_desktop_session_is_clipboard_enabled (RemoteDesktopSession *session)
+{
+  return session->clipboard.clipboard_enabled;
+}
+
+OrgGnomeMutterRemoteDesktopSession *
+remote_desktop_session_mutter_session_proxy (RemoteDesktopSession *session)
+{
+  return session->mutter_session_proxy;
+}
+
 static void
 remote_desktop_dialog_handle_free (RemoteDesktopDialogHandle *dialog_handle)
 {
@@ -160,6 +229,8 @@ remote_desktop_dialog_done (GtkWidget                 *widget,
                             int                        dialog_response,
                             RemoteDesktopDeviceType    device_types,
                             GPtrArray                 *streams,
+                            gboolean                   clipboard_enabled,
+                            gboolean                   persist_allowed,
                             RemoteDesktopDialogHandle *dialog_handle)
 {
   int response;
@@ -187,7 +258,12 @@ remote_desktop_dialog_done (GtkWidget                 *widget,
     {
       g_autoptr(GError) error = NULL;
 
-      if (!start_session (dialog_handle->session, device_types, streams, &error))
+      if (!start_session (dialog_handle->session,
+                          device_types,
+                          streams,
+                          clipboard_enabled,
+                          persist_allowed,
+                          &error))
         {
           g_warning ("Failed to start session: %s", error->message);
           response = 2;
@@ -229,7 +305,9 @@ create_remote_desktop_dialog (RemoteDesktopSession *session,
     GTK_WINDOW (remote_desktop_dialog_new (request->app_id,
                                            session->select.device_types,
                                            session->select.screen_cast_enable ?
-                                             &session->select.screen_cast : NULL));
+                                             &session->select.screen_cast : NULL,
+                                           session->clipboard.clipboard_requested,
+                                           session->persist_mode));
   gtk_window_set_transient_for (dialog, GTK_WINDOW (fake_parent));
   gtk_window_set_modal (dialog, TRUE);
 
@@ -266,6 +344,7 @@ on_mutter_session_closed (OrgGnomeMutterRemoteDesktopSession *mutter_session_pro
 static RemoteDesktopSession *
 remote_desktop_session_new (const char *app_id,
                             const char *session_handle,
+                            const char *peer_name,
                             const char *mutter_session_path,
                             OrgGnomeMutterRemoteDesktopSession *mutter_session_proxy)
 {
@@ -273,6 +352,7 @@ remote_desktop_session_new (const char *app_id,
 
   remote_desktop_session = g_object_new (remote_desktop_session_get_type (),
                                          "id", session_handle,
+                                         "peer-name", peer_name,
                                          NULL);
   remote_desktop_session->mutter_session_path =
     g_strdup (mutter_session_path);
@@ -334,6 +414,7 @@ handle_create_session (XdpImplRemoteDesktop *object,
 
   session = (Session *)remote_desktop_session_new (arg_app_id,
                                                    arg_session_handle,
+                                                   sender,
                                                    mutter_session_path,
                                                    mutter_session_proxy);
 
@@ -371,6 +452,8 @@ handle_select_devices (XdpImplRemoteDesktop *object,
                        const char *arg_app_id,
                        GVariant *arg_options)
 {
+  g_autofree char *provider = NULL;
+  g_autoptr(GVariant) restore_data = NULL;
   const char *sender;
   g_autoptr(Request) request = NULL;
   RemoteDesktopSession *remote_desktop_session;
@@ -378,6 +461,7 @@ handle_select_devices (XdpImplRemoteDesktop *object,
   uint32_t device_types;
   GVariantBuilder results_builder;
   GVariant *results;
+  uint32_t version;
 
   sender = g_dbus_method_invocation_get_sender (invocation);
   request = request_new (sender, arg_app_id, arg_handle);
@@ -397,6 +481,25 @@ handle_select_devices (XdpImplRemoteDesktop *object,
   remote_desktop_session->select.device_types = device_types;
   response = 0;
 
+  g_variant_lookup (arg_options, "persist_mode", "u",
+                    &remote_desktop_session->persist_mode);
+
+  if (g_variant_lookup (arg_options, "restore_data", "(suv)",
+                        &provider, &version, &restore_data))
+    {
+      if (!g_variant_check_format_string (restore_data,
+                                          REMOTE_DESKTOP_RESTORE_VARIANT_TYPE,
+                                          FALSE))
+        {
+          g_warning ("Cannot parse restore data, ignoring");
+          goto out;
+        }
+
+      if (g_strcmp0 (provider, "GNOME") == 0 &&
+          version == RESTORE_FORMAT_VERSION)
+        remote_desktop_session->restored.data = g_variant_ref (restore_data);
+    }
+
 out:
   request_export (request, g_dbus_method_invocation_get_connection (invocation));
 
@@ -408,12 +511,51 @@ out:
   return TRUE;
 }
 
+static GVariant *
+serialize_session_as_restore_data (RemoteDesktopSession *remote_desktop_session,
+                                   GPtrArray            *streams)
+{
+  GVariantBuilder restore_data_builder;
+  GVariantBuilder impl_builder;
+  int64_t creation_time;
+  int64_t last_used_time;
+
+  if ((!streams || streams->len == 0) &&
+      !remote_desktop_session->shared.device_types)
+    return NULL;
+
+  last_used_time = g_get_real_time ();
+  if (remote_desktop_session->restored.creation_time != -1)
+    creation_time = remote_desktop_session->restored.creation_time;
+  else
+    creation_time = g_get_real_time ();
+
+  g_variant_builder_init (&impl_builder,
+                          G_VARIANT_TYPE (REMOTE_DESKTOP_RESTORE_VARIANT_TYPE));
+  g_variant_builder_add (&impl_builder, "x", creation_time);
+  g_variant_builder_add (&impl_builder, "x", last_used_time);
+
+  g_variant_builder_add (&impl_builder, "u",
+                         remote_desktop_session->shared.device_types);
+  g_variant_builder_add (&impl_builder, "b",
+                         remote_desktop_session->clipboard.clipboard_enabled);
+  serialize_screen_cast_streams_as_restore_data (streams, &impl_builder);
+
+  g_variant_builder_init (&restore_data_builder, G_VARIANT_TYPE ("(suv)"));
+  g_variant_builder_add (&restore_data_builder, "s", "GNOME");
+  g_variant_builder_add (&restore_data_builder, "u", RESTORE_FORMAT_VERSION);
+  g_variant_builder_add (&restore_data_builder, "v", g_variant_builder_end (&impl_builder));
+  return g_variant_builder_end (&restore_data_builder);
+}
+
 static void
 start_done (RemoteDesktopSession *remote_desktop_session)
 {
   GVariantBuilder results_builder;
   RemoteDesktopDeviceType shared_device_types;
   GnomeScreenCastSession *gnome_screen_cast_session;
+  gboolean clipboard_enabled;
+  RemoteDesktopPersistMode persist_mode;
 
   g_variant_builder_init (&results_builder, G_VARIANT_TYPE_VARDICT);
 
@@ -421,6 +563,37 @@ start_done (RemoteDesktopSession *remote_desktop_session)
   g_variant_builder_add (&results_builder, "{sv}",
                          "devices", g_variant_new_uint32 (shared_device_types));
 
+  if (remote_desktop_session->clipboard.clipboard_requested)
+  {
+    g_autoptr (GError) error = NULL;
+    GVariantBuilder options_builder;
+    GVariant *options;
+
+    remote_desktop_session->clipboard.selection_owner_changed_handler_id =
+      g_signal_connect (remote_desktop_session->mutter_session_proxy,
+                        "selection-owner-changed",
+                        G_CALLBACK (on_selection_owner_changed),
+                        remote_desktop_session);
+    remote_desktop_session->clipboard.selection_transfer_handler_id =
+      g_signal_connect (remote_desktop_session->mutter_session_proxy,
+                        "selection-transfer",
+                        G_CALLBACK (on_selection_transfer),
+                        remote_desktop_session);
+
+    g_variant_builder_init (&options_builder, G_VARIANT_TYPE_VARDICT);
+    options = g_variant_builder_end (&options_builder);
+
+    if (!org_gnome_mutter_remote_desktop_session_call_enable_clipboard_sync (
+      remote_desktop_session->mutter_session_proxy, options, NULL, &error))
+      {
+        g_warning ("Failed to enable clipboard: %s", error->message);
+      }
+
+    clipboard_enabled = remote_desktop_session->clipboard.clipboard_enabled;
+    g_variant_builder_add (&results_builder, "{sv}", "clipboard_enabled",
+                           g_variant_new_boolean (clipboard_enabled));
+  }
+
   gnome_screen_cast_session = remote_desktop_session->gnome_screen_cast_session;
   if (gnome_screen_cast_session)
     {
@@ -434,6 +607,26 @@ start_done (RemoteDesktopSession *remote_desktop_session)
                              g_variant_builder_end (&streams_builder));
     }
 
+  persist_mode = remote_desktop_session->persist_mode;
+  if (persist_mode != REMOTE_DESKTOP_PERSIST_MODE_NONE &&
+      remote_desktop_session->shared.persist_allowed)
+    {
+      g_autoptr(GPtrArray) streams = NULL;
+      GVariant *restore_data;
+
+      streams = g_steal_pointer (&remote_desktop_session->streams_to_restore);
+      restore_data = serialize_session_as_restore_data (remote_desktop_session,
+                                                        streams);
+
+      if (restore_data)
+        {
+          g_variant_builder_add (&results_builder, "{sv}", "persist_mode",
+                                 g_variant_new_uint32 (persist_mode));
+          g_variant_builder_add (&results_builder, "{sv}", "restore_data",
+                                 restore_data);
+        }
+    }
+
   xdp_impl_remote_desktop_complete_start (XDP_IMPL_REMOTE_DESKTOP (impl),
                                           remote_desktop_session->start_invocation,
                                           0,
@@ -473,6 +666,9 @@ open_screen_cast_session (RemoteDesktopSession  *remote_desktop_session,
                       G_CALLBACK (on_gnome_screen_cast_session_ready),
                       remote_desktop_session);
 
+  if (remote_desktop_session->persist_mode != REMOTE_DESKTOP_PERSIST_MODE_NONE)
+    remote_desktop_session->streams_to_restore = g_ptr_array_ref (streams);
+
   if (!gnome_screen_cast_session_record_selections (gnome_screen_cast_session,
                                                     streams,
                                                     &remote_desktop_session->select.screen_cast,
@@ -486,12 +682,16 @@ static gboolean
 start_session (RemoteDesktopSession     *remote_desktop_session,
                RemoteDesktopDeviceType   device_types,
                GPtrArray                *streams,
+               gboolean                  clipboard_enabled,
+               gboolean                  persist_allowed,
                GError                  **error)
 {
   OrgGnomeMutterRemoteDesktopSession *session_proxy;
   gboolean need_streams;
 
+  remote_desktop_session->shared.persist_allowed = persist_allowed;
   remote_desktop_session->shared.device_types = device_types;
+  remote_desktop_session->clipboard.clipboard_enabled = clipboard_enabled;
 
   if (streams)
     {
@@ -530,6 +730,53 @@ cancel_start_session (RemoteDesktopSession *remote_desktop_session,
                                           g_variant_builder_end (&results_builder));
 }
 
+static gboolean
+restore_from_data (RemoteDesktopSession *remote_desktop_session)
+{
+  g_autoptr(GPtrArray) streams = NULL;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(GVariantIter) screen_cast_streams_iter = NULL;
+  int64_t creation_time;
+  int64_t last_used_time;
+  uint32_t restored_device_types;
+  gboolean clipboard_enabled;
+
+  if (!remote_desktop_session->restored.data)
+    return FALSE;
+
+  g_variant_get (remote_desktop_session->restored.data,
+                 REMOTE_DESKTOP_RESTORE_VARIANT_TYPE,
+                 &creation_time,
+                 &last_used_time,
+                 &restored_device_types,
+                 &clipboard_enabled,
+                 &screen_cast_streams_iter);
+
+  if (restored_device_types & ~REMOTE_DESKTOP_DEVICE_TYPE_ALL)
+    {
+      g_warning ("Tried to restore bogus device type mask 0x%x",
+                 restored_device_types);
+      return FALSE;
+    }
+
+  streams = restore_screen_cast_streams (screen_cast_streams_iter,
+                                         &remote_desktop_session->select.screen_cast);
+
+  remote_desktop_session->restored.creation_time = creation_time;
+
+  if (!start_session (remote_desktop_session,
+                      restored_device_types, streams,
+                      clipboard_enabled,
+                      TRUE,
+                      &error))
+    {
+      g_warning ("Error restoring stream from session: %s", error->message);
+      return FALSE;
+    }
+
+  return TRUE;
+}
+
 static gboolean
 handle_start (XdpImplRemoteDesktop *object,
               GDBusMethodInvocation *invocation,
@@ -542,7 +789,6 @@ handle_start (XdpImplRemoteDesktop *object,
   const char *sender;
   g_autoptr(Request) request = NULL;
   RemoteDesktopSession *remote_desktop_session;
-  RemoteDesktopDialogHandle *dialog_handle;
   GVariantBuilder results_builder;
 
   sender = g_dbus_method_invocation_get_sender (invocation);
@@ -564,13 +810,18 @@ handle_start (XdpImplRemoteDesktop *object,
       goto err;
     }
 
-  dialog_handle = create_remote_desktop_dialog (remote_desktop_session,
-                                                invocation,
-                                                request,
-                                                arg_parent_window);
-
   remote_desktop_session->start_invocation = invocation;
-  remote_desktop_session->dialog_handle = dialog_handle;
+
+  if (!restore_from_data (remote_desktop_session))
+    {
+      RemoteDesktopDialogHandle *dialog_handle;
+
+      dialog_handle = create_remote_desktop_dialog (remote_desktop_session,
+                                                    invocation,
+                                                    request,
+                                                    arg_parent_window);
+      remote_desktop_session->dialog_handle = dialog_handle;
+    }
 
   return TRUE;
 
@@ -844,6 +1095,53 @@ handle_notify_touch_up (XdpImplRemoteDesktop *object,
   return TRUE;
 }
 
+static gboolean
+handle_connect_to_eis (XdpImplRemoteDesktop *object,
+                       GDBusMethodInvocation *invocation,
+                       const char *arg_app_id,
+                       const char *arg_session_handle,
+                       GVariant *arg_options)
+{
+  RemoteDesktopSession *remote_desktop_session;
+  OrgGnomeMutterRemoteDesktopSession *proxy;
+  g_autoptr(GError) error = NULL;
+  g_autoptr(GUnixFDList) in_fd_list = NULL;
+  g_autoptr(GUnixFDList) out_fd_list = NULL;
+  GVariant *out_fd;
+  GVariantBuilder options_builder;
+  GVariant *options;
+
+  remote_desktop_session =
+    (RemoteDesktopSession *)lookup_session (arg_session_handle);
+  proxy = remote_desktop_session->mutter_session_proxy;
+
+  g_variant_builder_init (&options_builder, G_VARIANT_TYPE_VARDICT);
+  g_variant_builder_add (&options_builder, "device-types", "u",
+                         remote_desktop_session->shared.device_types);
+  options = g_variant_builder_end (&options_builder);
+
+  if (!org_gnome_mutter_remote_desktop_session_call_connect_to_eis_sync (proxy,
+                                                                         options,
+                                                                         in_fd_list,
+                                                                         &out_fd,
+                                                                         &out_fd_list,
+                                                                         NULL,
+                                                                         &error))
+    {
+      g_warning ("Failed to connect to EIS: %s", error->message);
+      g_dbus_method_invocation_return_error (invocation,
+                                             XDG_DESKTOP_PORTAL_ERROR,
+                                             XDG_DESKTOP_PORTAL_ERROR_FAILED,
+                                             "Failed to connect to EIS");
+
+      return G_DBUS_METHOD_INVOCATION_HANDLED;
+    }
+
+  xdp_impl_remote_desktop_complete_connect_to_eis (object, invocation, out_fd_list, out_fd);
+
+  return G_DBUS_METHOD_INVOCATION_HANDLED;
+}
+
 static unsigned int
 gnome_device_types_xdp_device_types (unsigned int gnome_device_types)
 {
@@ -883,6 +1181,7 @@ remote_desktop_name_appeared (GDBusConnection *connection,
     }
 
   impl = G_DBUS_INTERFACE_SKELETON (xdp_impl_remote_desktop_skeleton_new ());
+  xdp_impl_remote_desktop_set_version (XDP_IMPL_REMOTE_DESKTOP (impl), 2);
 
   g_signal_connect (impl, "handle-create-session",
                     G_CALLBACK (handle_create_session), NULL);
@@ -911,6 +1210,8 @@ remote_desktop_name_appeared (GDBusConnection *connection,
                     G_CALLBACK (handle_notify_touch_motion), NULL);
   g_signal_connect (impl, "handle-notify-touch-up",
                     G_CALLBACK (handle_notify_touch_up), NULL);
+  g_signal_connect (impl, "handle-connect-to-eis",
+                    G_CALLBACK (handle_connect_to_eis), NULL);
 
   supported_device_types =
     org_gnome_mutter_remote_desktop_get_supported_device_types (remote_desktop);
@@ -984,6 +1285,16 @@ remote_desktop_session_close (Session *session)
   g_signal_handler_disconnect (session_proxy,
                                remote_desktop_session->closed_handler_id);
 
+  if (remote_desktop_session->clipboard.clipboard_enabled)
+    {
+      g_signal_handler_disconnect (
+          session_proxy, remote_desktop_session->clipboard
+                             .selection_owner_changed_handler_id);
+      g_signal_handler_disconnect (
+          session_proxy, remote_desktop_session->clipboard
+                             .selection_transfer_handler_id);
+    }
+
   if (!org_gnome_mutter_remote_desktop_session_call_stop_sync (session_proxy,
                                                                NULL,
                                                                &error))
@@ -998,6 +1309,9 @@ remote_desktop_session_finalize (GObject *object)
 {
   RemoteDesktopSession *remote_desktop_session = (RemoteDesktopSession *)object;
 
+  g_clear_pointer (&remote_desktop_session->streams_to_restore,
+                   g_ptr_array_unref);
+  g_clear_pointer (&remote_desktop_session->restored.data, g_variant_unref);
   g_free (remote_desktop_session->mutter_session_path);
 
   G_OBJECT_CLASS (remote_desktop_session_parent_class)->finalize (object);
@@ -1019,4 +1333,25 @@ remote_desktop_session_class_init (RemoteDesktopSessionClass *klass)
 
   session_class = (SessionClass *)klass;
   session_class->close = remote_desktop_session_close;
+
+  signals[CLIPBOARD_SELECTION_OWNER_CHANGED] =
+    g_signal_new ("clipboard-selection-owner-changed",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE,
+                  1,
+                  G_TYPE_VARIANT);
+
+  signals[CLIPBOARD_SELECTION_TRANSFER] =
+    g_signal_new ("clipboard-selection-transfer",
+                  G_TYPE_FROM_CLASS (klass),
+                  G_SIGNAL_RUN_LAST,
+                  0,
+                  NULL, NULL, NULL,
+                  G_TYPE_NONE,
+                  2,
+                  G_TYPE_STRING,
+                  G_TYPE_INT);
 }
