@@ -31,6 +31,9 @@ struct _RemoteDesktopDialog
 
   GtkWidget *accept_button;
   GtkSwitch *allow_remote_interaction_switch;
+  GtkSwitch *allow_remote_clipboard_row;
+  GtkSwitch *allow_remote_clipboard_switch;
+  GtkCheckButton *persist_check;
   GtkWidget *screen_cast_widget;
   GtkHeaderBar *titlebar;
 
@@ -67,12 +70,20 @@ get_selected_device_types (RemoteDesktopDialog *dialog)
     return REMOTE_DESKTOP_DEVICE_TYPE_NONE;
 }
 
+static gboolean
+get_persist (RemoteDesktopDialog *dialog)
+{
+  return gtk_check_button_get_active (dialog->persist_check);
+}
+
 static void
 button_clicked (GtkWidget *button,
                 RemoteDesktopDialog *dialog)
 {
   RemoteDesktopDeviceType device_types = 0;
+  gboolean enable_clipboard;
   g_autoptr(GPtrArray) streams = NULL;
+  gboolean persist = FALSE;
   int response;
 
   gtk_widget_set_visible (GTK_WIDGET (dialog), FALSE);
@@ -84,17 +95,23 @@ button_clicked (GtkWidget *button,
 
       response = GTK_RESPONSE_OK;
       device_types = get_selected_device_types (dialog);
+
+      persist = get_persist (dialog);
+
       if (dialog->screen_cast_enable)
         streams = screen_cast_widget_get_selected_streams (screen_cast_widget);
+      enable_clipboard = gtk_switch_get_active (dialog->allow_remote_clipboard_switch);
     }
   else
     {
       response = GTK_RESPONSE_CANCEL;
       device_types = 0;
       streams = NULL;
+      enable_clipboard = FALSE;
     }
 
-  g_signal_emit (dialog, signals[DONE], 0, response, device_types, streams);
+  g_signal_emit (dialog, signals[DONE], 0, response,
+                 device_types, streams, enable_clipboard, persist);
 }
 
 static void
@@ -134,7 +151,9 @@ on_has_selection_changed (ScreenCastWidget *screen_cast_widget,
 RemoteDesktopDialog *
 remote_desktop_dialog_new (const char *app_id,
                            RemoteDesktopDeviceType device_types,
-                           ScreenCastSelection *screen_cast_select)
+                           ScreenCastSelection *screen_cast_select,
+                           gboolean clipboard_requested,
+                           RemoteDesktopPersistMode persist_mode)
 {
   RemoteDesktopDialog *dialog;
 
@@ -173,6 +192,31 @@ remote_desktop_dialog_new (const char *app_id,
 
     }
 
+  if (clipboard_requested)
+    {
+      gtk_widget_set_visible (GTK_WIDGET (dialog->allow_remote_clipboard_row),
+                              TRUE);
+      gtk_switch_set_active (dialog->allow_remote_clipboard_switch, TRUE);
+    }
+  else
+    {
+      gtk_widget_set_visible (GTK_WIDGET (dialog->allow_remote_clipboard_row),
+                              FALSE);
+      gtk_switch_set_active (dialog->allow_remote_clipboard_switch, FALSE);
+    }
+
+  switch (persist_mode)
+    {
+    case REMOTE_DESKTOP_PERSIST_MODE_NONE:
+      break;
+    case REMOTE_DESKTOP_PERSIST_MODE_PERSISTENT:
+      gtk_widget_set_visible (GTK_WIDGET (dialog->persist_check), TRUE);
+      G_GNUC_FALLTHROUGH;
+    case REMOTE_DESKTOP_PERSIST_MODE_TRANSIENT:
+      gtk_check_button_set_active (dialog->persist_check, TRUE);
+      break;
+    }
+
   return dialog;
 }
 
@@ -187,7 +231,7 @@ remote_desktop_dialog_close_request (GtkWindow *dialog)
 {
   gtk_widget_set_visible (GTK_WIDGET (dialog), FALSE);
 
-  g_signal_emit (dialog, signals[DONE], 0, GTK_RESPONSE_CANCEL, 0, NULL);
+  g_signal_emit (dialog, signals[DONE], 0, GTK_RESPONSE_CANCEL, 0, NULL, FALSE, FALSE);
 
   return TRUE;
 }
@@ -206,16 +250,21 @@ remote_desktop_dialog_class_init (RemoteDesktopDialogClass *klass)
                                 0,
                                 NULL, NULL,
                                 NULL,
-                                G_TYPE_NONE, 3,
+                                G_TYPE_NONE, 5,
                                 G_TYPE_INT,
                                 G_TYPE_INT,
-                                G_TYPE_PTR_ARRAY);
+                                G_TYPE_PTR_ARRAY,
+                                G_TYPE_BOOLEAN,
+                                G_TYPE_BOOLEAN);
 
   g_type_ensure (SCREEN_CAST_TYPE_WIDGET);
 
   gtk_widget_class_set_template_from_resource (widget_class, "/org/freedesktop/portal/desktop/gnome/remotedesktopdialog.ui");
   gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, accept_button);
   gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, allow_remote_interaction_switch);
+  gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, allow_remote_clipboard_row);
+  gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, allow_remote_clipboard_switch);
+  gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, persist_check);
   gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, screen_cast_widget);
   gtk_widget_class_bind_template_child (widget_class, RemoteDesktopDialog, titlebar);
   gtk_widget_class_bind_template_callback (widget_class, button_clicked);
